import request from 'superagent'
import Ajv from 'ajv'
import { version } from './package.json'
import md5 from 'md5'

const SERVER_URL = 'https://api.airshiphq.com'
const IDENTIFY_ENDPOINT = `${SERVER_URL}/v1/identify`
const GATING_INFO_ENDPOINT = `${SERVER_URL}/v1/gating-info`
const PLATFORM = 'nodejs'
const VERSION = version

const SDK_VERSION = `${PLATFORM}:${VERSION}`

const CONTROL_TYPE_BOOLEAN = 'boolean'
const CONTROL_TYPE_MULTIVARIATE = 'multivariate'

const DISTRIBUTION_TYPE_RULE_BASED = 'R'
const DISTRIBUTION_TYPE_PERCENTAGE_BASED = 'P'

const OBJECT_ATTRIBUTE_TYPE_STRING = 'STRING'
const OBJECT_ATTRIBUTE_TYPE_INT = 'INT'
const OBJECT_ATTRIBUTE_TYPE_FLOAT = 'FLOAT'
const OBJECT_ATTRIBUTE_TYPE_BOOLEAN = 'BOOLEAN'
const OBJECT_ATTRIBUTE_TYPE_DATE = 'DATE'
const OBJECT_ATTRIBUTE_TYPE_DATETIME = 'DATETIME'

const RULE_OPERATOR_TYPE_IS = 'IS'
const RULE_OPERATOR_TYPE_IS_NOT = 'IS_NOT'
const RULE_OPERATOR_TYPE_IN = 'IN'
const RULE_OPERATOR_TYPE_NOT_IN = 'NOT_IN'
const RULE_OPERATOR_TYPE_LT = 'LT'
const RULE_OPERATOR_TYPE_LTE = 'LTE'
const RULE_OPERATOR_TYPE_GT = 'GT'
const RULE_OPERATOR_TYPE_GTE = 'GTE'
const RULE_OPERATOR_TYPE_FROM = 'FROM'
const RULE_OPERATOR_TYPE_UNTIL = 'UNTIL'
const RULE_OPERATOR_TYPE_AFTER = 'AFTER'
const RULE_OPERATOR_TYPE_BEFORE = 'BEFORE'

const SCHEMA = {
  "type": "object",
  "properties": {
    "type": {
      "type": "string",
      "pattern": "^([A-Z][a-zA-Z]*)+$",
      "maxLength": 50,
    },
    "isGroup": {
      "type": "boolean",
    },
    "id": {
      "type": ["string", "integer"],
      "pattern": "^[ -~]+$",
      "maxLength": 250,
      "minLength": 1,
    },
    "displayName": {
      "type": "string",
      "maxLength": 250,
      "minLength": 1,
    },
    "isAnonymous": {
      "type": "boolean",
    },
    "attributes": {
      "type": "object",
      "patternProperties": {
        "^[a-zA-Z][a-zA-Z_]{0,48}[a-zA-Z]$": {
          "oneOf": [
            {
              "type": "string",
              "maxLength": 3000,
            },
            {
              "type": "boolean"
            },
            {
              "type": "number"
            },
          ],
        },
      },
      "maxProperties": 100,
      "additionalProperties": false,
    },
    "group": {
      "type": ["object", "null"],
      "properties": {
        "type": {
          "type": "string",
          "pattern": "^([A-Z][a-zA-Z]*)+$",
          "maxLength": 50,
        },
        "isGroup": {
          "type": "boolean",
          "enum": [true],
        },
        "id": {
          "type": ["string", "integer"],
          "pattern": "^[ -~]+$",
          "maxLength": 250,
          "minLength": 1,
        },
        "displayName": {
          "type": "string",
          "maxLength": 250,
          "minLength": 1,
        },
        "attributes": {
          "type": "object",
          "patternProperties": {
            "^[a-zA-Z][a-zA-Z_]{0,48}[a-zA-Z]$": {
              "oneOf": [
                {
                  "type": "string",
                  "maxLength": 3000,
                },
                {
                  "type": "boolean"
                },
                {
                  "type": "number"
                },
              ],
            },
          },
          "maxProperties": 100,
          "additionalProperties": false,
        },
      },
      "required": ["id"],
      "additionalProperties": false,
    },
  },
  "required": ["id"],
  "additionalProperties": false,
}

const makeid = () => {
  let text = ''
  let possible = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

  for (let i = 0; i < 6; i++) {
    text += possible.charAt(Math.floor(Math.random() * possible.length))
  }

  return text
}

const SDK_ID = makeid()


const getHashedValue = (s) => {
  return parseInt(md5(s), 16) * 1.0 / 340282366920938463463374607431768211455
}

class Airship {
  constructor(options) {
    this.apiKey = options.apiKey
    this.envKey = options.envKey
    this.timeout = 10 * 1000

    this.pollingInterval = 60 * 1000
    this.ingestionInterval = 60 * 1000

    this.pollingThread = null
    this.ingestionThread = null

    this.gatingInfo = null
    this.gatingInfoMap = null

    this.ingestionBatch = []
    this.ingestionMaxBatchSize = 500

    let ajv = Ajv()
    this.validate = ajv.compile(SCHEMA)
    this.firstIngestion = true
  }

  init = () => {
    this.ingestionThread = setInterval(this.triggerIngestion, this.ingestionInterval)

    let getGatingInfo = () => {
      return request.get(`${GATING_INFO_ENDPOINT}/${this.envKey}`)
        .set('api-key', this.apiKey)
        .set('sdk-version', SDK_VERSION)
        .timeout(this.timeout)
    }

    let maybeGetGatingInfo = () => {
      let gatingInfoPromise = getGatingInfo().then(res => {
        let gatingInfo = res.body

        if (gatingInfo.serverInfo === 'maintenance') {
          console.error('Airship is currently going through maintenance')
        } else {
          let gatingInfoMap = this._getGatingInfoMap(gatingInfo)

          this.gatingInfo = gatingInfo
          this.gatingInfoMap = gatingInfoMap

          if (gatingInfo.pollingInterval) {
            let newPollingInterval = gatingInfo.pollingInterval * 1000

            if (this.pollingInterval !== newPollingInterval) {
              clearInterval(this.pollingThread)

              this.pollingInterval = newPollingInterval
              this.pollingThread = setInterval(maybeGetGatingInfo, this.pollingInterval)
            }
          }

          if (gatingInfo.ingestionInterval) {
            let newIngestionInterval = gatingInfo.ingestionInterval * 1000

            if (this.ingestionInterval !== newIngestionInterval) {
              clearInterval(this.ingestionThread)

              this.ingestionInterval = newIngestionInterval
              this.ingestionThread = setInterval(this.triggerIngestion, this.ingestionInterval)
            }
          }
        }
      }).catch(err => {
        console.error('Failed to connect to Airship server')
      })
      return gatingInfoPromise
    }

    let initialGatingInfoPromise = maybeGetGatingInfo()

    this.pollingThread = setInterval(maybeGetGatingInfo, this.pollingInterval)

    return initialGatingInfoPromise
  }

  triggerIngestion = () => {
    if (this.ingestionBatch.length === 0) {
      return
    }

    let doIngestion = () => {
      let payload = this.ingestionBatch
      this.ingestionBatch = []

      return request.post(IDENTIFY_ENDPOINT)
        .type('application/json')
        .set('api-key', this.apiKey)
        .timeout(this.timeout)
        .send({
          envKey: this.envKey,
          objects: payload,
        }).then(() => {}).catch(() => {})
    }

    return doIngestion()
  }

  _getGatingInfoMap = (gatingInfo) => {
    let map = {}

    let controls = gatingInfo.controls

    for (let i = 0; i < controls.length; i++) {
      let control = controls[i]
      let controlInfo = {}

      controlInfo.id = control.id
      controlInfo.isOn = control.isOn
      controlInfo.ruleBasedDistributionDefaultVariation = control.ruleBasedDistributionDefaultVariation
      controlInfo.ruleSets = control.ruleSets
      controlInfo.distributions = control.distributions
      controlInfo.type = control.type
      controlInfo.defaultVariation = control.defaultVariation

      let enablements = control.enablements
      let enablementsInfo = {}

      for (let j = 0; j < enablements.length; j++) {
        let enablement = enablements[j]

        let clientIdentitiesMap = enablementsInfo[enablement.clientObjectTypeName]

        if (clientIdentitiesMap === undefined) {
          enablementsInfo[enablement.clientObjectTypeName] = {}
        }

        enablementsInfo[enablement.clientObjectTypeName][enablement.clientObjectIdentity] = [enablement.isEnabled, enablement.variation]
      }

      controlInfo.enablementsInfo = enablementsInfo

      map[control.shortName] = controlInfo
    }

    return map
  }

  _ingestAsync = (gateStats) => {
    this.ingestionBatch.push(gateStats)

    if (this.firstIngestion) {
      setImmediate(this.triggerIngestion)
      this.firstIngestion = false
      return
    }

    if (this.ingestionBatch.length >= this.ingestionMaxBatchSize) {
      setImmediate(this.triggerIngestion)
    }
  }

  _satisfiesRule = (rule, object) => {
    let attributeType = rule.attributeType
    let operator = rule.operator
    let attributeName = rule.attributeName
    let value = rule.value
    let valueList = rule.valueList

    if (object.attributes === undefined || object.attributes[attributeName] === undefined) {
      return false
    }

    let attributeVal = object.attributes[attributeName]

    if (attributeType === OBJECT_ATTRIBUTE_TYPE_STRING) {
      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else if (operator === RULE_OPERATOR_TYPE_IN) {
        return valueList.indexOf(attributeVal) >= 0
      } else if (operator === RULE_OPERATOR_TYPE_NOT_IN) {
        return valueList.indexOf(attributeVal) === -1
      } else {
        return false
      }
    } else if (attributeType === OBJECT_ATTRIBUTE_TYPE_INT) {
      value = value && parseInt(value)
      valueList = valueList && valueList.map(v => parseInt(v))

      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else if (operator === RULE_OPERATOR_TYPE_IN) {
        return valueList.indexOf(attributeVal) >= 0
      } else if (operator === RULE_OPERATOR_TYPE_NOT_IN) {
        return valueList.indexOf(attributeVal) === -1
      } else if (operator === RULE_OPERATOR_TYPE_LT) {
        return attributeVal < value
      } else if (operator === RULE_OPERATOR_TYPE_LTE) {
        return attributeVal <= value
      } else if (operator === RULE_OPERATOR_TYPE_GT) {
        return attributeVal > value
      } else if (operator === RULE_OPERATOR_TYPE_GTE) {
        return attributeVal >= value
      } else {
        return false
      }
    } else if (attributeType === OBJECT_ATTRIBUTE_TYPE_FLOAT) {
      value = value && parseFloat(value)
      valueList = valueList && valueList.map(v => parseFloat(v))

      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else if (operator === RULE_OPERATOR_TYPE_IN) {
        return valueList.indexOf(attributeVal) >= 0
      } else if (operator === RULE_OPERATOR_TYPE_NOT_IN) {
        return valueList.indexOf(attributeVal) === -1
      } else if (operator === RULE_OPERATOR_TYPE_LT) {
        return attributeVal < value
      } else if (operator === RULE_OPERATOR_TYPE_LTE) {
        return attributeVal <= value
      } else if (operator === RULE_OPERATOR_TYPE_GT) {
        return attributeVal > value
      } else if (operator === RULE_OPERATOR_TYPE_GTE) {
        return attributeVal >= value
      } else {
        return false
      }
    } else if (attributeType === OBJECT_ATTRIBUTE_TYPE_BOOLEAN) {
      value = (value === 'true')
      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else {
        return false
      }
    } else if (attributeType === OBJECT_ATTRIBUTE_TYPE_DATE) {
      let unixTimestamp = (new Date(attributeVal)).getTime()

      if (isNaN(unixTimestamp)) {
        return false
      }

      let isoFormat = (new Date(attributeVal)).toISOString()

      if (!isoFormat.endsWith('T00:00:00.000Z')) {
        return false
      }

      value = value && (new Date(value)).getTime()
      valueList = valueList && valueList.map(v => (new Date(v)).getTime())

      attributeVal = unixTimestamp

      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else if (operator === RULE_OPERATOR_TYPE_IN) {
        return valueList.indexOf(attributeVal) >= 0
      } else if (operator === RULE_OPERATOR_TYPE_NOT_IN) {
        return valueList.indexOf(attributeVal) === -1
      } else if (operator === RULE_OPERATOR_TYPE_FROM) {
        return attributeVal >= value
      } else if (operator === RULE_OPERATOR_TYPE_UNTIL) {
        return attributeVal <= value
      } else if (operator === RULE_OPERATOR_TYPE_AFTER) {
        return attributeVal > value
      } else if (operator === RULE_OPERATOR_TYPE_BEFORE) {
        return attributeVal < value
      } else {
        return false
      }
    } else if (attributeType === OBJECT_ATTRIBUTE_TYPE_DATETIME) {
      // .getTime() respects timezones
      let unixTimestamp = (new Date(attributeVal)).getTime()

      if (isNaN(unixTimestamp)) {
        return false
      }

      value = value && (new Date(value)).getTime()
      valueList = valueList && valueList.map(v => (new Date(v)).getTime())

      attributeVal = unixTimestamp

      if (operator === RULE_OPERATOR_TYPE_IS) {
        return attributeVal === value
      } else if (operator === RULE_OPERATOR_TYPE_IS_NOT) {
        return attributeVal !== value
      } else if (operator === RULE_OPERATOR_TYPE_IN) {
        return valueList.indexOf(attributeVal) >= 0
      } else if (operator === RULE_OPERATOR_TYPE_NOT_IN) {
        return valueList.indexOf(attributeVal) === -1
      } else if (operator === RULE_OPERATOR_TYPE_FROM) {
        return attributeVal >= value
      } else if (operator === RULE_OPERATOR_TYPE_UNTIL) {
        return attributeVal <= value
      } else if (operator === RULE_OPERATOR_TYPE_AFTER) {
        return attributeVal > value
      } else if (operator === RULE_OPERATOR_TYPE_BEFORE) {
        return attributeVal < value
      } else {
        return false
      }
    } else {
      return false
    }
  }

  _getGateValuesForObject = (controlInfo, object) => {
    if (controlInfo.enablementsInfo[object.type] !== undefined) {
      if (controlInfo.enablementsInfo[object.type][object.id] !== undefined) {
        let [isEnabled, variation] = controlInfo.enablementsInfo[object.type][object.id]
        return {
          isEnabled,
          variation,
          isEligible: isEnabled,
          _fromEnablement: true,
        }
      }
    }

    let sampledInsideBasePopulation = false
    let isEligible = false
    for (let i = 0; i < controlInfo.ruleSets.length; i++) {
      if (sampledInsideBasePopulation) {
        break
      }
      let ruleSet = controlInfo.ruleSets[i]
      let rules = ruleSet.rules

      if (ruleSet.clientObjectTypeName !== object.type) {
        continue
      }

      let satisfiesAllRules = true
      for (let j = 0; j < rules.length; j++) {
        let rule = rules[j]
        satisfiesAllRules = satisfiesAllRules && this._satisfiesRule(rule, object)
      }

      if (satisfiesAllRules) {
        isEligible = true
        let hashKey = `SAMPLING:control_${controlInfo.id}:env_${this.gatingInfo.env.id}:rule_set_${ruleSet.id}:client_object_${object.type}_${object.id}`
        if (getHashedValue(hashKey) <= ruleSet.samplingPercentage) {
          sampledInsideBasePopulation = true
        }
      }
    }

    if (!sampledInsideBasePopulation) {
      return {
        isEnabled: false,
        variation: null,
        isEligible: isEligible,
      }
    }

    if (controlInfo.type === CONTROL_TYPE_BOOLEAN) {
      return {
        isEnabled: true,
        variation: null,
        isEligible: true,
      }
    } else if (controlInfo.type === CONTROL_TYPE_MULTIVARIATE) {
      if (controlInfo.distributions.length === 0) {
        return {
          isEnabled: true,
          variation: controlInfo.defaultVariation,
          isEligible: true,
        }
      }

      let percentageBasedDistributions = controlInfo.distributions.filter(d => d.type === DISTRIBUTION_TYPE_PERCENTAGE_BASED)
      let ruleBasedDistributions = controlInfo.distributions.filter(d => d.type === DISTRIBUTION_TYPE_RULE_BASED)

      if (percentageBasedDistributions.length !== 0 && ruleBasedDistributions.length !== 0) {
        console.error('Rule integrity error: please contact support@airshiphq.com')
        return {
          isEnabled: false,
          variation: null,
          isEligible: false,
        }
      }

      if (percentageBasedDistributions.length !== 0) {
        let delta = 0.0001
        let sum_percentages = 0.0
        let running_percentages = []
        for (let i = 0; i < percentageBasedDistributions.length; i++) {
          let distribution = percentageBasedDistributions[i]
          sum_percentages += distribution.percentage
          if (running_percentages.length === 0) {
            running_percentages.push(distribution.percentage)
          } else {
            running_percentages.push(running_percentages[running_percentages.length - 1] + distribution.percentage)
          }
        }

        if (Math.abs(1.0 - sum_percentages) > delta) {
          console.error('Rule integrity error: please contact support@airshiphq.com')
          return {
            isEnabled: false,
            variation: null,
            isEligible: false,
          }
        }

        let hashKey = `DISTRIBUTION:control_${controlInfo.id}:env_${this.gatingInfo.env.id}:client_object_${object.type}_${object.id}`
        let hashedPercentage = getHashedValue(hashKey)

        for (let i = 0; i < running_percentages.length; i++) {
          let percentage = running_percentages[i]
          if (hashedPercentage <= percentage) {
            return {
              isEnabled: true,
              variation: percentageBasedDistributions[i].variation,
              isEligible: true,
            }
          }
        }

        return {
          isEnabled: true,
          variation: percentageBasedDistributions[percentageBasedDistributions.length - 1].variation,
          isEligible: true,
        }
      } else {
        for (let i = 0; i < ruleBasedDistributions.length; i++) {
          let distribution = ruleBasedDistributions[i]

          let ruleSet = distribution.ruleSet
          let rules = ruleSet.rules

          if (ruleSet.clientObjectTypeName !== object.type) {
            continue
          }

          let satisfiesAllRules = true
          for (let j = 0; j < rules.length; j++) {
            let rule = rules[j]
            satisfiesAllRules = satisfiesAllRules && this._satisfiesRule(rule, object)
          }

          if (satisfiesAllRules) {
            return {
              isEnabled: true,
              variation: distribution.variation,
              isEligible: true,
            }
          }
        }

        return {
          isEnabled: true,
          variation: controlInfo.ruleBasedDistributionDefaultVariation || controlInfo.defaultVariation,
          isEligible: true,
          _ruleBasedDefaultVariation: true,
        }
      }
    } else {
      return {
        isEnabled: false,
        variation: null,
        isEligible: false,
      }
    }
  }

  _getGateValues = (controlShortName, object) => {
    if (this.gatingInfoMap[controlShortName] === undefined) {
      return {
        isEnabled: false,
        variation: null,
        isEligible: false,
        _shouldSendStats: false,
      }
    }

    let controlInfo = this.gatingInfoMap[controlShortName]

    if (!controlInfo.isOn) {
      return {
        isEnabled: false,
        variation: null,
        isEligible: false,
        _shouldSendStats: true,
      }
    }

    let group = null
    if (object.group !== undefined) {
      group = object.group
    }

    let result = this._getGateValuesForObject(controlInfo, object)

    if (group !== null) {
      if (group.type === undefined) {
        group.type = `${object.type}Group`
        group.isGroup = true
      }
      let groupResult = this._getGateValuesForObject(controlInfo, group)

      if (result._fromEnablement && !result.isEnabled) {
        // Do nothing
      } else if (!result._fromEnablement && groupResult._fromEnablement && !groupResult.isEnabled) {
        result.isEnabled = groupResult.isEnabled
        result.variation = groupResult.variation
        result.isEligible = groupResult.isEligible
      } else if (result.isEnabled) {
        if (result._ruleBasedDefaultVariation) {
          if (groupResult.isEnabled) {
            result.isEnabled = groupResult.isEnabled
            result.variation = groupResult.variation
            result.isEligible = groupResult.isEligible
          } else {
            // Do nothing
          }
        } else {
          // Do nothing
        }
      } else if (groupResult.isEnabled) {
        result.isEnabled = groupResult.isEnabled
        result.variation = groupResult.variation
        result.isEligible = groupResult.isEligible
      } else {
        // Do nothing
      }
    }

    result._shouldSendStats = true
    return result
  }

  _cloneObject = (object) => {
    let clone = Object.assign({}, object)

    if (object.attributes !== undefined) {
      clone.attributes = Object.assign({}, object.attributes)
    }

    if (object.group !== undefined) {
      clone.group = Object.assign({}, object.group)

      if (object.group.attributes !== undefined) {
        clone.group.attributes = Object.assign({}, object.group.attributes)
      }
    }

    return clone
  }

  _validateNesting = (object) => {
    if (object.isGroup === true && object.group !== undefined) {
      return 'A group cannot be nested inside another group'
    }

    return null
  }

  _validateAnonymous = (object) => {
    if (object.isAnonymous === true && object.group !== undefined) {
      return 'An anonymous object cannot belong to a group'
    }

    if (object.isGroup === true && object.isAnonymous === true) {
      return 'A group cannot be anonymous'
    }

    return null
  }

  _maybeTransformId = (object) => {
    if (Number.isInteger(object.id)) {
      let idStr = '' + object.id
      if (idStr.length > 250) {
        return 'Integer id must have 250 digits or less'
      }
      object.id = idStr
    }

    let group = null
    if (object.group !== undefined) {
      group = object.group
    }

    if (group !== null) {
      if (Number.isInteger(group.id)) {
        let idStr = '' + group.id
        if (idStr.length > 250) {
          return 'Integer id must have 250 digits or less'
        }
        group.id = idStr
      }
    }

    return null
  }

  _enrichWithMetadata = (controlShortName, stats) => {
    let controlInfo = this.gatingInfoMap[controlShortName]

    if (controlInfo !== undefined) {
      stats.sdkGateControlId = controlInfo.id
    }

    stats.sdkEnvId = this.gatingInfo['env'].id
  }

  _maybeAddFields = (object) => {
    if (object.type === undefined) {
      object.type = 'User'
    }

    if (object.displayName === undefined) {
      object.displayName = '' + object.id
    }

    let group = null
    if (object.group !== undefined) {
      group = object.group
    }

    if (group !== null && group.displayName === undefined) {
      group.displayName = '' + group.id
    }
  }

  isEnabled = (controlShortName, object, defaultValue=false) => {
    let valid = this.validate(object)

    if (!valid) {
      console.error(this.validate.errors)
      return defaultValue
    }

    object = this._cloneObject(object)

    this._maybeAddFields(object)

    let error = this._validateNesting(object) || this._validateAnonymous(object) || this._maybeTransformId(object)

    if (error) {
      console.error(error)
      return defaultValue
    }

    if (this.gatingInfoMap === null) {
      return defaultValue
    }

    let gateTimestamp = (new Date()).toISOString()

    let start = process.hrtime()
    let { isEnabled, variation, isEligible, _shouldSendStats } = this._getGateValues(controlShortName, object)
    let end = process.hrtime(start)

    if (_shouldSendStats) {
      let sdkGateTimestamp = gateTimestamp
      let sdkGateLatency = `${end[1] / 1000.0}us`
      let sdkVersion = SDK_VERSION

      let stats = {}
      stats.sdkGateControlShortName = controlShortName
      stats.sdkGateTimestamp = sdkGateTimestamp
      stats.sdkGateLatency = sdkGateLatency

      stats.sdkGateValue = isEnabled
      stats.sdkGateVariation = variation
      stats.sdkGateEligibility = isEligible
      stats.sdkGateType = 'value'

      this._enrichWithMetadata(controlShortName, stats)

      stats.sdkVersion = sdkVersion
      stats.sdkId = SDK_ID

      object.stats = stats

      this._ingestAsync(object)
    }

    return isEnabled
  }

  getVariation = (controlShortName, object, defaultValue=null) => {
    let valid = this.validate(object)

    if (!valid) {
      console.error(this.validate.errors)
      return defaultValue
    }

    object = this._cloneObject(object)

    this._maybeAddFields(object)

    let error = this._validateNesting(object) || this._validateAnonymous(object) || this._maybeTransformId(object)

    if (error) {
      console.error(error)
      return defaultValue
    }

    if (this.gatingInfoMap === null) {
      return defaultValue
    }

    let gateTimestamp = (new Date()).toISOString()

    let start = process.hrtime()
    let { isEnabled, variation, isEligible, _shouldSendStats } = this._getGateValues(controlShortName, object)
    let end = process.hrtime(start)

    if (_shouldSendStats) {
      let sdkGateTimestamp = gateTimestamp
      let sdkGateLatency = `${end[1] / 1000.0}us`
      let sdkVersion = SDK_VERSION

      let stats = {}
      stats.sdkGateControlShortName = controlShortName
      stats.sdkGateTimestamp = sdkGateTimestamp
      stats.sdkGateLatency = sdkGateLatency

      stats.sdkGateValue = isEnabled
      stats.sdkGateVariation = variation
      stats.sdkGateEligibility = isEligible
      stats.sdkGateType = 'variation'

      this._enrichWithMetadata(controlShortName, stats)

      stats.sdkVersion = sdkVersion
      stats.sdkId = SDK_ID

      object.stats = stats

      this._ingestAsync(object)
    }

    return variation
  }

  isEligible = (controlShortName, object, defaultValue=false) => {
    let valid = this.validate(object)

    if (!valid) {
      console.error(this.validate.errors)
      return defaultValue
    }

    object = this._cloneObject(object)

    this._maybeAddFields(object)

    let error = this._validateNesting(object) || this._validateAnonymous(object) || this._maybeTransformId(object)

    if (error) {
      console.error(error)
      return defaultValue
    }

    if (this.gatingInfoMap === null) {
      return defaultValue
    }

    let gateTimestamp = (new Date()).toISOString()

    let start = process.hrtime()
    let { isEnabled, variation, isEligible, _shouldSendStats } = this._getGateValues(controlShortName, object)
    let end = process.hrtime(start)

    if (_shouldSendStats) {
      let sdkGateTimestamp = gateTimestamp
      let sdkGateLatency = `${end[1] / 1000.0}us`
      let sdkVersion = SDK_VERSION

      let stats = {}
      stats.sdkGateControlShortName = controlShortName
      stats.sdkGateTimestamp = sdkGateTimestamp
      stats.sdkGateLatency = sdkGateLatency

      stats.sdkGateValue = isEnabled
      stats.sdkGateVariation = variation
      stats.sdkGateEligibility = isEligible
      stats.sdkGateType = 'eligibility'

      this._enrichWithMetadata(controlShortName, stats)

      stats.sdkVersion = sdkVersion
      stats.sdkId = SDK_ID

      object.stats = stats

      this._ingestAsync(object)
    }

    return isEligible
  }
}

exports = module.exports = Airship
